/*
 * Copyright (C) 1997, 1998 Defence Research Establishment, Dept. of Comm.
 * Systems, FOA 72, Sweden.
 * Currently maintained by Christian Jnsson FOA 72 <chj@lin.foa.se>.
 *
 * This file is part of ETFTP version 1.1.3.
 *
 *   ETFTP is free software; you can redistribute it and/or modify it
 *   under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   ETFTP is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with ETFTP; see the file GPL.  If not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA 02111-1307, USA.
 *
 * Code:
 *
 *----------------------------------------------------------------------
 * Filenamne:   etftptime.c
 * Class:       Time
 * Created:     0.1, 970418, Per-Erik Forssn
 * Revised:     1.1, 980219, Torbjrn Frostemark
 *
 * Description: Keeps track of the time; serves as an abstraction of
 *              the system clock.
 * Remarks:     -
 *----------------------------------------------------------------------
 */

#include "etftptime.h"


/*----------------------------------------------------------------------
 *
 * TimeGet
 *
 *  This method asks the operating system for the current time.
 *
 * Results:
 *  Time in seconds.
 *
 * Side effects:
 *  None.
 *
 *----------------------------------------------------------------------
 */

long TimeGet()
{
  static long tptr;
  return time(&tptr);
}


/*----------------------------------------------------------------------
 *
 * TimeDate
 *
 *  This method asks the operating system for the current date string.
 *
 * Results:
 *  A string containing the current date.
 *
 * Side effects:
 *  None.
 *
 *----------------------------------------------------------------------
 */

char *TimeDate()
{
  long tsecs;
  static char *tstr;
  struct tm *tdate;

  tsecs = time(&tsecs);
  tdate = localtime(&tsecs);
  tstr = asctime(tdate);
  tstr[strlen(tstr)] = 000;
  tstr[strlen(tstr) - 1] = 000;
  return tstr;
}
