/* 
 * Copyright (C) 1997, 1998 Defence Research Establishment, Dept. of Comm.
 * Systems, FOA 72, Sweden.
 * Currently maintained by Christian Jnsson FOA 72 <chj@lin.foa.se>.
 *
 * This file is part of ETFTP version 1.1.3.
 *
 *   ETFTP is free software; you can redistribute it and/or modify it
 *   under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   ETFTP is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with ETFTP; see the file GPL.  If not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA 02111-1307, USA.
 *
 * Code:
 *
 *----------------------------------------------------------------------
 * Filenamne:	net.c
 * Class:	Net
 * Created:	0.1, 970423, Per-Erik Forssn
 * Revised:	1.1, 980219, Torbjrn Frostemark
 * Revised:	1.2, 980825, Christian Jnsson
 * 
 * Description: Interface to the UDP-layer.
 * Remarks:	-
 *
 * Private Metods:	-
 *
 * Private Attributes:	int sockfd
 *			char *openHost
 *			sockaddr_in UDPLocalAddr
 *			sockaddr_in UDPRemoteAddr
 *			servent UDPServInfo
 *			hostent UDPHostInfo
 *			int recvFirst
 *----------------------------------------------------------------------
 */

#ifdef SOLARIS_ETFTP
#include <sys/filio.h>
#endif /* SOLARIS_ETFTP */

#include "net.h"
#include "clu.h"
#include "etftptime.h"

static int  sockfd = -1;		/* fd for socket of server */
static char openHost[MAXHOSTNAME] = { 0 };	/* remember host's name */

extern int  errno;
extern int  traceFlag;			/* For the DEBUGn macros */
static int  recvFirst = 0;    /* Flag set if NetRecv should set port number */

int timeoutFlag = FALSE;

struct sockaddr_in	UDPLocalAddr;	/* local Internet socket addr */
struct sockaddr_in	UDPRemoteAddr;	/* remote Internet socket addr */
struct servent		UDPServInfo;	/* from getservbyname() */
struct hostent		UDPHostInfo;	/* from gethostbyname() */

int	h_errno;			/* host error number */
int	h_nerr;				/* number of error message strings */
char	*h_errlist[];			/* the error message table */

#ifdef SERVER
int  recvnBytes = -1;
extern char *recvPack;			/* Buffer (declared in init.c) */
#endif /* SERVER */

/*----------------------------------------------------------------------
 * NetInit
 *
 *	Creates a socket for server to listen on.
 *	Need not be called if we were started by inetd.
 *
 * Results:
 *	Returns -1 on failure 0 on success.
 *
 * Side effects:
 *	Initiates the socket.
 *----------------------------------------------------------------------
 */

#ifdef SERVER
int NetInit(service,port)
     char *service;	/* the name of the service (should be "ETFTP") */
     int port;		/* New port number or 0. */
{
  /*
   * Restore when etftp is in /etc/services
   * struct servent  *sp; 
   */
  DEBUG2("NetInit: Called with service = %s port = %d",service,port);
  /*
   * We weren't started by a master daemon.
   * We have to create a socket ourselves and bind our well-known
   * address to it.
   */

  if (port == 0) {
    port = 1818;   /* Default to this port (ugly and VERY temporary.) */
  }
  
  if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    DEBUG0("NetInit: Can't create datagram socket");
    return -1;
  }
  
  /*
   * Bind our local address so that any client can send to us.
   */
  memset(&UDPLocalAddr, 0, sizeof(struct sockaddr_in));
  

  UDPLocalAddr.sin_family      = AF_INET;
  UDPLocalAddr.sin_port        = htons(port);
  UDPLocalAddr.sin_addr.s_addr = htonl(INADDR_ANY);

  DEBUG1("NetInit: local socket is: %d", ntohs(UDPLocalAddr.sin_port));
  
  if (bind(sockfd, (struct sockaddr *)&UDPLocalAddr,
	   sizeof(UDPLocalAddr)) < 0)
    {
      DEBUG0("NetInit: can't bind local address");
      perror("etftp");
      return -1;
    }

  DEBUG1("NetInit: bind to socket %d complete", ntohs(UDPLocalAddr.sin_port)); 

  return 0;
}
#endif /* SERVER */

/*
 *----------------------------------------------------------------------
 * NetOpen
 *
 *	Open the network connection for a client.
 *
 * Results:
 *	Returns -1 on failure 0 on success.
 *	
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

#ifdef CLIENT
int ClientNetOpen(host,service,port)
     char *host;	/* name of other system to communicate with */
     char *service;	/* name of service (should be "ETFTP") */
     int  port; 	/* new port number or 0 */
{
  DEBUG3("ClientNetOpen: Called with host = %s service = %s port = %d",host,service,port);    

  /*
   * Call UDPOpen() to create the socket.  We tell UDPOpen to
   * not connect the socket, since we'll receive the first response
   * from a port that's different from where we send our first
   * datagram to.
   * FOR NOW no service is specified, when etftp is added in /etc/services
   * this should be the service.
   */
  
  if ((sockfd = ClientUDPOpen(host, NULL, port)) < 0) {
    return -1;
  }
  
  DEBUG2("ClientNetOpen: host %s, port %d",
	 inet_ntoa(UDPRemoteAddr.sin_addr),
	 ntohs(UDPRemoteAddr.sin_port));
  
  strcpy(openHost, host);	/* save the host's name */
  recvFirst = 1;		/* flag for NetRecv() */

  return 0;
}
#endif /* CLIENT */

/*----------------------------------------------------------------------
 * NetListen
 *
 *	Act server; listen for requests
 *      Listen for messages, then change to new port.
 *
 * Results:
 *	Process id for daughter if parent and 0 if child.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

#ifdef SERVER
/*
 * Initiate the server's end.
 * We are passed a flag that says whether or not we were started
 * by a "master daemon," such as the inetd program under 4.3BSD.
 * A master daemon will have already waited for a message to arrive
 * for us, and will have already set up the connection to the client.
 * If we weren't started by a master daemon, then we must wait for a
 * client's request to arrive.
 */

int NetListen(inetdflag)
    int inetdflag;	/* true if inetd started us */
{
  register int    childpid, nbytes;
  int 	    on;
  
  on = 1;
  DEBUG1("NetListen: inetdflag = %d",inetdflag);    
  if (inetdflag) {
    
    /*#ifdef	BSD		*//* assumes 4.3BSD inetd */
    /*    DEBUG0("BSD BSD BSD");*/
    /*
     * When we're fired up by inetd under 4.3BSD, file
     * descriptors 0, 1 and 2 are sockets to the client.
     * We want to first receive the message that's waiting
     * for us on the socket, and then close the socket.
     * This will let inetd go back to waiting for another
     * request on our "well-known port."
     */

    sockfd = 0;	/* descriptor for ServerNetRecv() to recvfrom() */

    /*
     * Set the socket to nonblocking, since inetd won't invoke
     * us unless there's a datagram ready for us to read.
     */

    if (ioctl(sockfd, FIONBIO, (char *) &on) < 0) {
      DEBUG0("NetListen:FIONBIO control test failed on local socket.");
      perror("etftp");
    }

    /*#endif*/	/* BSD inetd specifics */

  }

  /*
   * Now read the first message from the client.
   * In the inetd case, the message is already here and the call to
   * ServerNetRecv() returns immediately.  In the other case, ServerNetRecv()
   * blocks until a client request arrives.
   */

  recvFirst	=  1;	    /* tell NetRecv to save the address */
  recvnBytes = -1;	    /* tell NetRecv to do the actual read */
  
  nbytes = ServerNetRecv(recvPack, MAXBUFF);
  DEBUG1("NetListen: recvPack = %p", recvPack);
  /*
   * Fork a child process to handle the client's request.
   * In the inetd case, the parent exits, which allows inetd to
   * handle the next request that arrives to this well-known port
   * (inetd's wait mode for a datagram socket).
   * Otherwise the parent returns the child pid to the caller, which
   * is probably a concurrent server that'll call us again, to wait
   * for the next client request to this well-known port.
   */
  DEBUG0("Forking new server...");

  if ((childpid = fork()) < 0) {
    DEBUG0("NetListen: Server can't fork");
  }
  
  else if (childpid > 0) {    /* parent */
    if (inetdflag) {
      DEBUG0("Main server terminating. Child forked...");
      exit(0);	    /* inetd case; we're done */
    } else {
      return childpid;    /* independent server */
    }
  }

  /*
   * Child process continues here.
   * First close the socket that is bound to the well-known address:
   * the parent will handle any further requests that arrive there.
   * We've already read the message that arrived for us to handle.
   */
  DEBUG0("Child server: setting recvnBytes to -1");
  recvFirst = 0;

  if (inetdflag) {
    close(0);
    close(1);
    close(2);
  } else {
    close(sockfd);		/* close socket for well-known address */
  }
  errno = 0;		    /* in case it was set by a close() */
  
  /*
   * Create a new socket.
   * Bind any local port number to the socket as our local address.
   * We don't connect(), since NetSend() uses the sendto()
   * system call, specifying the destination address each time.
   */

  if ( (sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    DEBUG0("NetListen: can't create socket");
  }
  memset(&UDPLocalAddr,0,sizeof(struct sockaddr_in));
  
  UDPLocalAddr.sin_family      = AF_INET;
  UDPLocalAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  UDPLocalAddr.sin_port	       = htons(0);
  if (bind(sockfd, (struct sockaddr *)&UDPLocalAddr, sizeof(UDPLocalAddr)) < 0) {
    DEBUG0("NetListen: bind error");
  }
  DEBUG1("NetListen: New Server port: %d", ntohs(UDPLocalAddr.sin_port));

  /*
   * Now we'll set a special flag for ServerNetRecv(), so that
   * the next time it's called, it'll know that the recvPack[]
   * already has the received packet in it (from our call to
   * ServerNetRecv() above).
   */

  recvnBytes = nbytes;
  return 0;
}
#endif /* SERVER */

/*
 *----------------------------------------------------------------------
 * NetClose
 *
 *	Close the network connection.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Sets sockfd to -1.
 *----------------------------------------------------------------------
 */

void NetClose()
{
    DEBUG2("NetClose: host = %s, fd = %d", openHost, sockfd);
    close(sockfd);
    sockfd = -1;
}

/*
 *----------------------------------------------------------------------
 * NetSend
 *
 *	Send data using UDP to the other end.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

void NetSend(buff,len)
     char    *buff;
     int     len;
{
  register int    rc;
  static char tmpTime[24];
  /*int i = 0;*/
  DEBUG3("NetSend: sending %d bytes host %s, port %d",
	 len, inet_ntoa(UDPRemoteAddr.sin_addr),
	 ntohs(UDPRemoteAddr.sin_port));
  if ((int)PF_TYPE(buff) == PT_CONTROL) {
    DEBUG3("NetSend: Sending packet number %d (type %d, subtype %d)",
	   ++packetSCounter, (int)PF_TYPE(buff), (int)PF_SUBTYPE(buff));
  } else {
    DEBUG2("NetSend: Sending packet number %d (type %d)",
	   ++packetSCounter, (int)PF_TYPE(buff));
  }
  strcpy(tmpTime, TimeDate());
  DEBUG1("------------------------------------------- Sent: %s", tmpTime);
  /*for(i = 0; i < 11; i++) {*/
    /* Show the first 44 bytes in buffer */
  /*    DEBUG5( "%2d: %5d %5d %5d %5d", 4*i, ((unsigned char *)buff)[4*i], ((unsigned char *)buff)[4*i + 1],
	    ((unsigned char *)buff)[4*i + 2], ((unsigned char *)buff)[4*i + 3] );
    }*/

  rc = sendto(sockfd, buff, len, 0, (struct sockaddr *) &UDPRemoteAddr,
	      sizeof(struct sockaddr_in));
  if (rc != len) {
    DEBUG0("NetSend: Send error");
  }
}

/*
 *----------------------------------------------------------------------
 * ClientNetRecv
 *
 *	Recieve a record from the other end.
 *	The server version is allowed to timeout
 *	if timeoutFlag is set.   
 *
 * Results:
 *	-1 if interrupted(EINTR). Otherwise no of bytes in package.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

#ifdef CLIENT
int ClientNetRecv(buff,maxLen)
     char *buff;
     int maxLen;
{
  register int	nbytes;
  static char tmpTime[24];
#ifdef LINUX_ETFTP
  socklen_t 		fromlen;	/* value-result parameter */
#else
  int   fromlen;
#endif /* LINUX_ETFTP */
  struct sockaddr_in	from_addr;	/* actual addr of sender */
  fd_set fdvar;
  /*int i = 0;*/			/* debug */

  fromlen = sizeof(struct sockaddr_in);
  DEBUG0("ClientNetRecv: Waiting for packet...");
  
  FD_ZERO(&fdvar);
  FD_SET(sockfd, &fdvar);
  select(sockfd + 1, &fdvar, (fd_set *) 0, (fd_set *) 0, (struct timeval *) 0 ); 
  if (timeoutFlag) {
    DEBUG0("ClientNetRecv: Timed out...");
    return -1;
  } else {
    /*DEBUG0("ClientNetRecv: Incoming data present.");*/
    nbytes = recvfrom(sockfd, buff, maxLen, 0, (struct sockaddr *) &from_addr, &fromlen);
  }
    
  if (nbytes!=-1) {
    DEBUG3("ClientNetRecv: received %d bytes from %s, port %d",
	   nbytes, inet_ntoa(from_addr.sin_addr),
	   ntohs(from_addr.sin_port)); 

    DEBUG2("ClientNetRecv: Packet number %d (type %d) received.", ++packetRCounter, (int)PF_TYPE(recvPack));
    strcpy(tmpTime, TimeDate());
    DEBUG1("------------------------------------------- Received: %s", tmpTime);
    /*for(i = 0; i < 11; i++) {*/
      /* Show the first 44 bytes in packet */
    /* DEBUG5( "%2d: %5d %5d %5d %5d", 4*i, ((unsigned char *)buff)[4*i], ((unsigned char *)buff)[4*i + 1],
	      ((unsigned char *)buff)[4*i + 2], ((unsigned char *)buff)[4*i + 3] );
    }*/
  }
  
  /*
   * The recvfrom() system call can be interrupted by an alarm
   * interrupt, in case it times out.	We just return -1 if the
   * system call was interrupted, and the caller must determine
   * if this is OK or not.
   */

  if (nbytes < 0) {
    if (errno == EINTR) {
      return -1;
    } else {
      DEBUG0("ClientNetRecv: recvfrom error");
    }
  }

  /*
   * The TFTP client using UDP/IP has a funny requirement.
   * The problem is that UDP is being used for a
   * "connection-oriented" protocol, which it wasn't really
   * designed for.  Rather than tying up a single well-known
   * port number, the server changes its port after receiving
   * the first packet from a client.
   *
   * The first packet a client sends to the server (an RRQ or a WRQ)
   * must be sent to its well-known port number (69 for TFTP).
   * The server is then to choose some other port number for all
   * subsequent transfers.  The recvfrom() call above will return
   * the server's current address.  If the port number that we
   * sent the last packet to (udp_srv_addr.sin_port) is still equal to
   * the initial well-known port number (UDPServInfo.s_port), then we
   * must set the server's port for our next transmission to be
   * the port number from the recvfrom().  See Section 4 of the RFC
   * where it talks about the TID (= port number).
   *
   * Furthermore, after we have determined the port number that
   * we'll be receiving from, we can verify each datagram to make
   * certain its from the right place.
   */

  if (recvFirst) {
    /*
     * This is the first received message.
     * The server's port should have changed.
     */
      
    if (UDPRemoteAddr.sin_port == from_addr.sin_port) {
      DEBUG1("ClientNetRecv: First receive from port %d",
	     ntohs(from_addr.sin_port));
    }
    
    UDPRemoteAddr.sin_port = from_addr.sin_port;
    /* save the new port number of the server */
    recvFirst = 0;
    
  } else if (UDPRemoteAddr.sin_port != from_addr.sin_port) {
    DEBUG2("ClientNetRecv: Received from port %d, expected from port %d",
	   ntohs(from_addr.sin_port),
	   ntohs(UDPRemoteAddr.sin_port));
  }
  
  return nbytes ; /* return the actual length of the message */
}
#endif	/* CLIENT */

/*
 *----------------------------------------------------------------------
 * ServerNetRecv
 *
 *	Server version of NetRecv.
 *	Receive a record from the other end.
 *	We're called not only by the user, but also by NetListen() above,
 *	to read the first datagram after a "connection" is established.
 *
 * Results:
 *	-1 if interrupted(EINTR). Otherwise no of bytes in package.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

#ifdef SERVER
int ServerNetRecv(buff, maxLen)
     char    *buff;
     int     maxLen;
{
  register int	nbytes;
  static char tmpTime[24];
#ifdef LINUX_ETFTP
  socklen_t 		fromlen;	/* value-result parameter */
#else
  int   fromlen;
#endif /* LINUX_ETFTP */
  struct sockaddr_in	from_addr;
  /*int i;*/				/* debug */
  fd_set fdvar;

  /*  DEBUG2("ServerNetRecv: start receive with buff: %p. maxLen; %d.", buff, maxLen);*/
  if (recvnBytes >= 0) {
    /*
     * First message has been handled specially by NetListen().
     * It's already been read into recvPack[].
     */
    DEBUG1("recvnBytes = %d.", recvnBytes);
    nbytes = recvnBytes;
    recvnBytes = -1;
    DEBUG1("recvnBytes = %d.", recvnBytes);
    return nbytes;
  }

again:
  fromlen = sizeof(struct sockaddr_in);
  DEBUG0("ServerNetRecv: Waiting for packet...");

  if (recvFirst) {
    /* Parent will stop and wait for incoming data */
    memset(&from_addr,0,sizeof(struct sockaddr_in));
    nbytes = recvfrom(sockfd, buff, maxLen, 0, (struct sockaddr *) &from_addr, &fromlen);
  } else {
    FD_ZERO(&fdvar);
    FD_SET(sockfd, &fdvar);
    select(sockfd + 1, &fdvar, (fd_set *) 0, (fd_set *) 0, (struct timeval *) 0 ); 
    if (timeoutFlag) {
      DEBUG0("ServerNetRecv: Timed out...");
      return -1;
    } else {
      /*DEBUG0("ServerNetRecv: Incoming data present.");*/
      memset(&from_addr,0,sizeof(struct sockaddr_in));
      nbytes = recvfrom(sockfd, buff, maxLen, 0, (struct sockaddr *) &from_addr, &fromlen);
    }
  }
  /* Old stuff
  memset(&from_addr,0,sizeof(struct sockaddr_in));
  nbytes = recvfrom(sockfd, buff, maxLen, 0, (struct sockaddr *) &from_addr, &fromlen);
  */    

  if(from_addr.sin_family != AF_INET) {
    DEBUG0("ServerNetRecv: Non internet packet received.");
    perror("ServerNetRecv");
  }

  UDPLocalAddr.sin_family      = AF_INET;
  UDPLocalAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  UDPLocalAddr.sin_port        = htons(0);
  
  /*
   * The server can have its recvfrom() interrupted by either an
   * alarm timeout or by a SIGCLD interrupt.  If it's a timeout,
   * "timeoutFlag" will be set and we have to return to the caller
   * to let them determine if another receive should be initiated.
   * For a SIGCLD signal, we can restart the recvfrom() ourselves.
   */
  
  if (nbytes < 0) {
    if (errno == EINTR) {
      if (timeoutFlag) {
	DEBUG0("ServerNetRecv: Timeout!");
	return -1;
      }
      DEBUG0("ServerNetRecv: Retry...");
      errno = 0;	    /* assume SIGCLD */
      goto again;
    }
    /*    DEBUG0("ServerNetRecv: recvfrom error");*/
    
  }
  
  DEBUG3("ServerNetRecv: received %d bytes from %s, port %d",
	 nbytes, inet_ntoa(from_addr.sin_addr),
	 ntohs(from_addr.sin_port)); 
  if ((int)PF_TYPE(recvPack) == PT_CONTROL) {
    DEBUG4("ServerNetRecv: Packet number %d (type %d, subtype %d, seq.nr = %d) received.",
	   ++packetRCounter, (int)PF_TYPE(recvPack), (int)PF_SUBTYPE(recvPack), (int)PF_SEQNUM(recvPack));
  } else {
  DEBUG2("ServerNetRecv: Packet number %d (type %d) received.", ++packetRCounter, (int)PF_TYPE(recvPack));
  }
  strcpy(tmpTime, TimeDate());
  DEBUG1("------------------------------------------- Received: %s", tmpTime);
  /*for(i = 0; i < 11; i++) {*/
    /* Show the first 44 bytes in packet */
  /*DEBUG5( "%2d: %5d %5d %5d %5d", 4*i, ((unsigned char *)buff)[4*i], ((unsigned char *)buff)[4*i + 1],
	    ((unsigned char *)buff)[4*i + 2], ((unsigned char *)buff)[4*i + 3] );
  }*/

  
  /*
   * If "recvFirst" is set, then we must save the received
   * address that recvfrom() stored in "from_addr" in the
   * global "UDPRemoteAddr".
   */

  if (recvFirst) {
    memset(&UDPRemoteAddr,0,sizeof(struct sockaddr_in));
    memcpy(&UDPRemoteAddr,&from_addr,fromlen);
    DEBUG1("UDPremoteAddr = %d.", ntohs(UDPRemoteAddr.sin_port));
    recvFirst = 0;
  }

  /*
   * Make sure the message is from the expected client.
   */
  
  if (UDPRemoteAddr.sin_port != htons(0) &&
      UDPRemoteAddr.sin_port != from_addr.sin_port) {
    DEBUG2("ServerNetRecv: Received from port %d, expected from port %d",
	   ntohs(from_addr.sin_port), ntohs(UDPRemoteAddr.sin_port));
  }
  
  return nbytes;  /* return the actual length of the message */
}
#endif	/* SERVER */

#ifdef CLIENT
#ifndef	INADDR_NONE
#define	INADDR_NONE	0xffffffff	/* should be in <netinet/in.h> */
#endif /* INADDR_NONE */
#endif /* CLIENT */
/*
 *----------------------------------------------------------------------
 * ClientUDPOpen
 *
 *	Utility code to open a local socket for UDP 
 *	communication (for client).
 *
 * Results:
 *	local socket filedescriptor or -1 on error.
 *
 * Side effects:
 *	Initializes UDPRemoteAddr structure.
 *----------------------------------------------------------------------
 */

#ifdef CLIENT
int ClientUDPOpen(host, service, port)
     char	*host;		/* name of other system to communicate with */
     char	*service;	/* name of service being requested */
     /* can be NULL, if port > 0 */
     int	port;		/* new port number or 0 */
     /* if < 0, bind a local reserved port */
     /* if > 0, it's the port number of server (host-byte-order) */
{
  int		fd;
  unsigned long	inaddr;
  char		*HostErrStr();
  struct servent    *sp;
  struct hostent    *hp;
  if(service!=NULL) {
    DEBUG3("ClientUDPOpen: host = %s service = %s port = %d ",host,service,port);
  } else {
    DEBUG2("ClientUDPOpen: host = %s port = %d ",host,port);
  }
  /*
   * Initialize the server's Internet address structure.
   * We'll store the actual 4-byte Internet address and the
   * 2-byte port number below.
   */
  memset(&UDPRemoteAddr,0,sizeof(struct sockaddr_in));
  UDPRemoteAddr.sin_family = AF_INET;
  
  if (service != NULL) {
    if ( (sp = getservbyname(service, "udp")) == NULL) {
      DEBUG1("ClientUDPOpen: unknown service: %s/udp", service);
      return -1;
    }
    UDPServInfo = *sp;			/* structure copy */
    if (port > 0) {
      UDPRemoteAddr.sin_port = htons(port);
    /* caller's value */
    } else {
      UDPRemoteAddr.sin_port = sp->s_port;
    }
    /* service's value */
  } else {
    if (port <= 0) {
      DEBUG0("ClientUDPOpen: must specify either service or port");
      return -1;
    }
    UDPRemoteAddr.sin_port = htons(port);
  }
  
  /*
   * First try to convert the host name as a dotted-decimal number.
   * Only if that fails do we call gethostbyname().
   */
  
  if ( (inaddr = inet_addr(host)) != INADDR_NONE) {
    /* it's dotted-decimal */
    DEBUG0("ClientUDPOpen: Got host by inet_addr()");
    memcpy(&UDPRemoteAddr.sin_addr,&inaddr,sizeof(struct sockaddr_in));
    UDPHostInfo.h_name = NULL;
    
  } else {
    DEBUG0("ClientUDPOpen: Trying to get host by gethostbyname()");
    if ( (hp = gethostbyname(host)) == NULL) {
      DEBUG2("ClientUDPOpen: host name error: %s %s",
	      host, HostErrStr());
      return -1;
    }
    UDPHostInfo = *hp;	/* found it by name, structure copy */
    memcpy(&UDPRemoteAddr.sin_addr,hp->h_addr,hp->h_length);
  }
  
  if (port < 0) {
    DEBUG0("ClientUDPOpen: reserved ports not implemeneted yet");
  }  

  if ( (fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    DEBUG0("ClientUDPOpen: can't create UDP socket");
    return -1;
  }
  
  /*
   * Bind any local address for us.
   */

  memset(&UDPLocalAddr,0,sizeof(struct sockaddr_in));
  UDPLocalAddr.sin_family      = AF_INET;
  UDPLocalAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  UDPLocalAddr.sin_port        = htons(0);
  if (bind(fd, (struct sockaddr *) &UDPLocalAddr,
	   sizeof(struct sockaddr_in)) < 0) {
    DEBUG0("ClientUDPOpen: bind error");
    perror("etftp");
    close(fd);
    return -1;
  }
  
  return fd;
}
#endif	/* CLIENT */

/*
 *----------------------------------------------------------------------
 * HostErrStr
 *
 *	Return a string containing some additional information
 *	after a host name or address lookup error - gethostbyname() or
 *	gethostbyaddr().
 *
 * Results:
 *	socket filedescriptor or -1 on error.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

#ifdef CLIENT
char *HostErrStr()
{
  static char	msgstr[200];
  
  if (h_errno != 0) {
    if (h_errno > 0 && h_errno < h_nerr) {
      sprintf(msgstr, "(%s)", h_errlist[h_errno]);
    } else {
      sprintf(msgstr, "(h_errno = %d)", h_errno);
    }
  } else {
    msgstr[0] = '\0';
  }
  
  return msgstr;
}
#endif	/* CLIENT */

/*    E N D   O F   L I N E    */



